% Wavelet packet ICA example script on signals (1D) 

% External packages needed:
% FastICA
% BSICA available at http://appsrv.cse.cuhk.edu.hk/~kzhang/research.html

% MATLAB Toolboxes needed:
% Signal Processing Toolbox
% Wavelet Toolbox

%%%%%%%%%%%%%%%%%%%%%
% USER VARIABLES
%%%%%%%%%%%%%%%%%%%%%

% sine interference frequency (0-pi)
fint=0.41;

% choose the nonlinearity of the FastICA ('pow3', 'tanh', 'gaus')
nonlinearity = 'tanh';

% choose the number of levels of the 1D wavelet packet decomposition
level = 5;

% threshold for the WP reconstruction
threshold = 0.05;

%%%%%%%%%%%%%%%%%%%%%%%
% end of USER VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%

% set the MATLAB path
wd = pwd; p = strcat(wd, '\wp'); path(path, p);

% prepare the 1D signals
t = [0:9999]; f = [0:4999]/5000*pi; wind = [401:800];
h1 = figure; subplot(5,1,1);
h2 = figure; subplot(5,1,1);

s1 = sin(0.9*t).*sin(t/30);  % AM signal
S1 = fft(s1);
figure(h1); subplot(5,1,1); plot(t(wind), s1(wind), 'k'); ylabel('s_{1,I}');  title('Source signals, dependent component')
figure(h2); subplot(5,1,1); plot(f, abs(S1(1:5000)),'k'); axis([0 3.2 0 210]); title('Source signals spectra')

s2 = [ones(1,54) -ones(1, 50)]; s2 = [s2 s2 s2 s2 s2 s2 s2 s2 s2 s2]; % square wave
s2 = wkeep([s2 s2 s2 s2 s2 s2 s2 s2 s2 s2], 10000); 
S2 = fft(s2);
figure(h1); subplot(5,1,2); plot(t(wind), s2(wind), 'k'); ylabel('s_{2,I}'); axis([401 800 -1.3 1.3]);
figure(h2); subplot(5,1,2); plot(f, abs(S2(1:5000)),'k'); axis([0 3.2 0 420]);

% s3 = diff(randn(1, 10001)); save data/s3 s3 % HP noise
load data/s3
S3 = fft(s3);
figure(h1); subplot(5,1,3); plot(t(wind), s3(wind), 'k'); ylabel('s_{3,I}');
figure(h2); subplot(5,1,3); plot(f, abs(S3(1:5000)),'k'); axis([0 3.2 0 400]);

load data/voice; s4 = (circshift(glas, -1600))'; % voice
S4 = fft(s4);
figure(h1); subplot(5,1,4); plot(t(wind), s4(wind-100), 'k'); ylabel('s_{4,I}');
figure(h2); subplot(5,1,4); plot(f, abs(S4(1:5000)), 'k'); axis([0 3.2 0 700]);

s5 = 1.6*sin(t*fint); % dependent component
S5 = fft(s5);
figure(h1); subplot(5,1,5); plot(t(wind), s5(wind), 'k'); ylabel('wave of\newline   s_{i,D}'); xlabel('t'); axis([401 800 -2 2]);
figure(h2); subplot(5,1,5); plot(f, abs(S5(1:5000)),'k'); axis([0 3.2 0 590]); xlabel('\Omega (rad)');

% adding dependent component with random phase shift
s1 = s1 + 1.6*sin(t*fint - 0.0881); s2 = s2 + 1.6*sin(t*fint + 2.4586);
s3 = s3 + 1.6*sin(t*fint + 1.6468); s4 = s4 + 1.6*sin(t*fint - 0.2735);

% show 
h25 = figure; subplot(4,1,1);
figure(h25); subplot(4,1,1); plot(t(wind), s1(wind), 'k'); ylabel('s_1'), title('Source signals with dependent component added')
figure(h25); subplot(4,1,2); plot(t(wind), s2(wind), 'k'); ylabel('s_2')
figure(h25); subplot(4,1,3); plot(t(wind), s3(wind), 'k'); ylabel('s_3')
figure(h25); subplot(4,1,4); plot(t(wind), s4(wind), 'k'); ylabel('s_4'),  xlabel('t')

% form mixed signals ("observed" data)
S = [s1; s2; s3; s4];
A = [0.7 0.7 1 1.2; 1 1.4 1 0.3; 1.2 0.3 0.7 1.0; 0.4 1.1 1.2 0.6]; % matrica mijesanja
X = A * S;

x1 = X(1,:); x2 = X(2,:); x3 = X(3,:); x4 = X(4,:);
X1 = fft(x1); X2 = fft(x2); X3 = fft(x3); X4 = fft(x4); 

% show the mixtures ("observed" data)
h3 = figure; subplot(4,1,1);
figure(h3); subplot(4,1,1); plot(t(wind), x1(wind), 'k'); ylabel('x_1'), title('Observed signals')
figure(h3); subplot(4,1,2); plot(t(wind), x2(wind), 'k'); ylabel('x_2')
figure(h3); subplot(4,1,3); plot(t(wind), x3(wind), 'k'); ylabel('x_3')
figure(h3); subplot(4,1,4); plot(t(wind), x4(wind), 'k'); ylabel('x_4'),  xlabel('t')

h4 = figure; subplot(4,1,1);
figure(h4); subplot(4,1,1); plot(f, abs(X1(1:5000)), 'k'); axis([0 3.2 0 10000]);, title('Observed signals spectra')
figure(h4); subplot(4,1,2); plot(f, abs(X2(1:5000)), 'k'); axis([0 3.2 0 10000]);
figure(h4); subplot(4,1,3); plot(f, abs(X3(1:5000)), 'k'); axis([0 3.2 0 10000]);
figure(h4); subplot(4,1,4); plot(f, abs(X4(1:5000)), 'k'); axis([0 3.2 0 10000]); xlabel('\Omega (rad)');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% direct application of the ICA method
[Aa, W] = fastica(X,'numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');
Y = W*X;

ae_direct = amari_error(W*A)

% extract the recovered signals
y1 = Y(1,:); y2 = Y(2,:); y3 = Y(3,:); y4 = Y(4,:); 

% show
h5 = figure; subplot(4,1,1);
figure(h5); subplot(4,1,1); plot(t(wind), y1(wind), 'k');, title('Directly recovered signals')
figure(h5); subplot(4,1,2); plot(t(wind), y2(wind), 'k');
figure(h5); subplot(4,1,3); plot(t(wind), y3(wind), 'k');
figure(h5); subplot(4,1,4); plot(t(wind), y4(wind), 'k');  xlabel('t')

disp('Directly recovered signals. Press enter to continue...')
pause

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Innovation approach
inord=10;
t0=clock;
a1=lpc(x1,inord); a2=lpc(x2,inord); a3=lpc(x3,inord); a4=lpc(x4,inord);
a=(a1+a2+a3+a4)/4;
x1i=filter(a,1,x1); x2i=filter(a,1,x2); 
x3i=filter(a,1,x3); x4i=filter(a,1,x4); 

[Aainv, Winv] = fastica([x1i; x2i; x3i; x4i],'numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');
Yinv = Winv*X;
time_inn=etime(clock,t0)
ae_inn = amari_error(Winv*A)

disp('Innovation approach has ended. Press enter to continue...')
pause

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LP-HP filter approach (first level WP transform)
t0=clock;

[swp1,lp] = swpdec(x1, 1, 'sym8'); swp2 = swpdec(x2, 1, 'sym8');
swp3 = swpdec(x3, 1, 'sym8'); swp4 = swpdec(x4, 1, 'sym8');
a1 = swp1(2,:); a2 = swp2(2,:); a3 = swp3(2,:); a4 = swp4(2,:);     % take HP part as being more independent
[Ahp, Whp] = fastica([a1(:), a2(:), a3(:), a4(:)]','numOfIC', 4, 'approach', 'symm', 'g', nonlinearity,'displayMode', 'off');
Yhp = Whp*X;

ae_hp = amari_error(Whp*A)
time_hp=etime(clock,t0)

disp('HP filtering approach has ended. Press enter to continue.')
pause
disp('Starting WPSBICA approach...')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1D wavelet packets (WP) approach
t0=clock;

[swp1,lp] = swpdec(x1, level, 'sym8'); swp2 = swpdec(x2, level, 'sym8');
     swp3 = swpdec(x3, level, 'sym8'); swp4 = swpdec(x4, level, 'sym8');

lsum = size(lp,1);
MI = zeros(lsum+1, 1);

MI(1) = MI_measure(X); % MI of the input mixture

% MI of the wavelet packet subbands
for i = 1:lsum,
    a1 = swp1(i,:); a2 = swp2(i,:); a3 = swp3(i,:); a4 = swp4(i,:);
    MI(i+1) = MI_measure([a1(:), a2(:), a3(:), a4(:)]'); 
end

% normalization
MI = MI / max(MI);

% WP reconstruction from independent subbands
nlev=2^level;
lp = zeros(size(lp));
lp(end-nlev+1:end) = 1;

ind = find(MI(end-nlev+1:end) > threshold);
ind = nlev+ind-1;

lp(ind-1) = 0;
xc1 = swprec(swp1,lp,'sym8');  xc2 = swprec(swp2,lp,'sym8');
xc3 = swprec(swp3,lp,'sym8');  xc4 = swprec(swp4,lp,'sym8');
Xc = [xc1; xc2; xc3; xc4];

[Aacwp, Wcwp] = fastica(Xc,'numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');

Ycwp = Wcwp*Xc;
time_wcp=etime(clock,t0)

ae_wcp = amari_error(Wcwp*A)

% show the MI tree
format short g
showMI1(MI)

% extract the recovered signals
ycwp1 = Ycwp(1,:); ycwp2 = Ycwp(2,:); ycwp3 = Ycwp(3,:); ycwp4 = Ycwp(4,:); 

% show
h7 = figure; subplot(4,1,1);
figure(h7); subplot(4,1,1); plot(t(wind), ycwp1(wind), 'k'); ylabel('y_1'), title('WPSDICA recovered signals')
figure(h7); subplot(4,1,2); plot(t(wind), ycwp2(wind), 'k'); ylabel('y_2')
figure(h7); subplot(4,1,3); plot(t(wind), ycwp3(wind), 'k'); ylabel('y_3')
figure(h7); subplot(4,1,4); plot(t(wind), ycwp4(wind), 'k'); ylabel('y_4'), xlabel('t')

YCWP1 = fft(ycwp1); YCWP2 = fft(ycwp2); YCWP3 = fft(ycwp3); YCWP4 = fft(ycwp4); 

h8 = figure; subplot(4,1,1);
figure(h8); subplot(4,1,1); plot(f, abs(YCWP1(1:5000)), 'k'); axis([0 3.2 0 700]);, title('WPSDICA recovered signals spectra')
figure(h8); subplot(4,1,2); plot(f, abs(YCWP2(1:5000)), 'k'); axis([0 3.2 0 500]);
figure(h8); subplot(4,1,3); plot(f, abs(YCWP3(1:5000)), 'k'); axis([0 3.2 0 300]);
figure(h8); subplot(4,1,4); plot(f, abs(YCWP4(1:5000)), 'k'); axis([0 3.2 0 400]); xlabel('\Omega (rad)');

disp('WPSDICA approach has ended. Press enter to continue.')
pause
disp('Starting Zhan-Chan BSICA approach...')

% Zhang - Chan algorithm
[yI, y, hh, Wzc] = bsica(X, 17, A);
ae_Chan = amari_error(Wzc*A)